import logging
import traceback
from http import HTTPStatus
from typing import Any, List, Type

import httpx
import orjson
import uvicorn
from fastapi import APIRouter, FastAPI, HTTPException, Request
from fastapi.responses import ORJSONResponse
from kami import KamiClient
from loguru import logger
from pydantic import BaseModel
from typing import Callable


from .exceptions import InvalidSignatureException
from .middleware import SignatureMiddleware, ZstdMiddleware
from .types import InterceptHandler, PydanticModel, ServerHandlerFunc
from .utils import create_response

router = APIRouter()


class Server:
    def __init__(
        self,
        app: FastAPI | None = None,
        kami: KamiClient | None = None,
        log_level: str = None,  # type: ignore
    ) -> None:
        if not log_level:
            log_level = "INFO"
        assert log_level in [
            "INFO",
            "DEBUG",
            "TRACE",
            "SUCCESS",
            "WARNING",
            "ERROR",
            "CRITICAL",
        ]
        self._log_level = log_level

        self._configure_loguru_logging()

        self.app = app or FastAPI()
        self.kami = kami or KamiClient()
        self.app.include_router(router)
        self.app.add_middleware(ZstdMiddleware)
        self.app.add_middleware(SignatureMiddleware, kami=self.kami)
        # NOTE: here we register some exception handlers that make it easier to
        # write miner's code
        self._add_http_exception_handler()
        self._add_invalid_signature_exception_handler()
        self.add_global_exception_handler()
        self.config = None

    def _configure_loguru_logging(self) -> None:
        """Configure FastAPI/uvicorn to use loguru logging"""
        # Remove existing handlers from root logger
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        intercept_handler = InterceptHandler()

        # Intercept standard logging at root level
        level = logging.getLevelName(self._log_level)
        # NOTE: using this at debug and below gets extremely spammy
        logging.basicConfig(handlers=[intercept_handler], level=level)

        # Ensure specific loggers propagate to root logger
        loggers = (
            "uvicorn",
            "uvicorn.access",
            "uvicorn.error",
            "fastapi",
            "asyncio",
            "starlette",
        )
        for logger_name in loggers:
            logging_logger = logging.getLogger(logger_name)
            logging_logger.handlers = []
            logging_logger.propagate = True

    async def close(self):
        if hasattr(self, "server") and self.server:
            self.server.should_exit = True
            await self.server.shutdown()
        await self.kami.close()

    def add_global_exception_handler(self) -> None:
        """Register exception handlers to standardize error responses"""

        @self.app.exception_handler(Exception)
        async def global_exception_handler(  # pyright: ignore[reportUnusedFunction]
            request: Request, exc: Exception
        ) -> ORJSONResponse:
            """Convert all exceptions to standardized response format"""
            logger.error(f"Global exception: {str(exc)}", exc_info=True)
            if hasattr(exc, "status_code"):
                status_code = exc.status_code  # type: ignore
            else:
                status_code = 500

            return create_response(
                error=str(exc),  # Use exc, not exc.__cause__
                body={},
                status_code=status_code,  # Pass int, not exception object
            )

    def _add_http_exception_handler(self) -> None:
        """Register exception handlers to standardize error responses"""

        @self.app.exception_handler(HTTPException)
        async def http_exception_handler(  # pyright: ignore[reportUnusedFunction]
            request: Request, exc: HTTPException
        ) -> ORJSONResponse:
            """Convert HTTPExceptions to standardized response format"""
            logger.error(f"HTTPException: {str(exc)}")
            return create_response(
                error=str(exc.detail), body={}, status_code=exc.status_code
            )

    def _add_invalid_signature_exception_handler(self) -> None:
        """Register invalid signature exception handler to standardize error responses"""

        @self.app.exception_handler(InvalidSignatureException)
        async def invalid_signature_exc_handler(  # pyright: ignore[reportUnusedFunction]
            request: Request, exc: InvalidSignatureException
        ) -> ORJSONResponse:
            """Convert InvalidSignatureException to standardized response format"""
            logger.error(f"HTTPException: {str(exc)}")
            return create_response(
                error=str(exc),
                body={},
                status_code=HTTPStatus.FORBIDDEN,
            )

    def serve_synapse(
        self, synapse: Type[PydanticModel], handler: ServerHandlerFunc[PydanticModel]
    ) -> None:
        # NOTE: we always want to have signature middleware, as miners should
        # only be reachable by validators
        self.app = _register_route_handler(self.app, handler, model=synapse)

    async def initialise(self, port: int) -> bool:
        try:
            logger.info("Starting FastAPI server with uvicorn...")
            logger.info(f"Server will support the following routes: {self.app.routes=}")
            # NOTE: prevent uvicorn from overriding logging settings
            server_config = uvicorn.Config(
                app=self.app,
                host="0.0.0.0",
                port=port,
                workers=1,
                log_config=None,
                log_level=None,
                reload=False,
            )
            self.config = server_config
            logger.info(
                f"Using server config host:{server_config.host}, port: {server_config.port}, log_level: {server_config.log_level}"
            )
            self.server = uvicorn.Server(server_config)
            await self.server.serve()

            return True
        except Exception as e:
            logger.error(f"Error starting server: {str(e)}")
            return False

    async def get_external_ip(self):
        services: list[tuple[str, Callable]] = [
            ("https://checkip.amazonaws.com", lambda r: r.text.strip()),
            ("https://api.ipify.org", lambda r: r.text.strip()),
            ("https://icanhazip.com", lambda r: r.text.strip()),
            ("https://ifconfig.me", lambda r: r.text.strip()),
            ("https://httpbin.org/ip", lambda r: r.json()["origin"]),
        ]

        async with httpx.AsyncClient() as client:
            for url, parser in services:
                try:
                    response = await client.get(url, timeout=5)
                    return parser(response)
                except Exception:
                    continue
            raise Exception("All IP detection services failed")


def _register_route_handler(
    app: FastAPI,
    handler: ServerHandlerFunc[PydanticModel],
    model: Type[PydanticModel],
    # NOTE: let's just default to post for now
    methods: List[str] = ["POST", "HEAD"],
) -> FastAPI:
    """Register a route with a Pydantic model to allow easily adding new endpoints"""

    async def handler_wrapper(request: Request) -> ORJSONResponse:
        """Wrapper around the request that handles zstd decompression and payload validation"""
        try:
            if request.method == "HEAD":
                return create_response(status_code=HTTPStatus.OK, body={})

            data: dict[str, Any] = {}
            try:
                # NOTE: we should be able to just read the data directly since
                # there's ZstdMiddleware enabled
                data = orjson.loads(await request.body())
            except orjson.JSONDecodeError as e:
                logger.error(f"JSON Decode error: {str(e)}")
                return create_response(
                    error=f"Invalid JSON, exception: {str(e)}",
                    body=data,
                    status_code=400,
                )

            try:
                logger.debug(f"Attempting to validate payload: {data=}")
                payload = model.model_validate(data)
            except Exception as e:
                logger.error(f"Validation error: {str(e)}")
                return create_response(
                    error=f"Validation error: {str(e)}",
                    body=data,
                    status_code=400,
                )

            # TODO: figure out why result is None?
            result = await handler(request, payload)

            logger.success(
                f"Handler: {handler.__name__}, result type:{type(result)}, result:{result}"
            )
            if not isinstance(result, dict):
                if isinstance(result, bytes):
                    result = orjson.loads(result)
                elif issubclass(type(result), BaseModel):
                    result = result.model_dump()
                return create_response(body=result)

            return create_response(body=result)
        except HTTPException as e:
            logger.error(f"HTTPException: {str(e)}")
            raise e
        except Exception as e:
            traceback.print_exc()
            logger.error(f"Error processing request due to: {str(e)}")
            # TODO: add more context here
            return create_response(error=f"Internal server error: {str(e)}", body={})

    # grab docstrings from underlying function
    description = handler.__doc__ if handler.__doc__ else handler_wrapper.__doc__
    handler_wrapper.__name__ = handler.__name__
    app.add_api_route(
        path="/" + model.__name__.lstrip("/").rstrip("/"),
        endpoint=handler_wrapper,
        methods=methods,
        operation_id=f"handle_{model.__name__}",
        description=description,
    )
    return app



class InvalidSignatureException(Exception):
    """Exception raised when signature is invalid"""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)



import logging
import traceback
from typing import Any, Awaitable, Callable, Generic, TypeAlias, TypeVar

import aiohttp
from fastapi import Request
from loguru import logger
from pydantic import BaseModel, ConfigDict, field_serializer

PydanticModel = TypeVar("PydanticModel", bound=BaseModel)
# define a pydantic model here so that we can apply these to child of BaseModel
ServerHandlerFunc: TypeAlias = Callable[[Request, PydanticModel], Awaitable[Any]]

SIGNATURE_HEADER = "x-signature"
HOTKEY_HEADER = "x-hotkey"
MESSAGE_HEADER = "x-message"


class StdResponse(BaseModel, Generic[PydanticModel]):
    """Standardized response that preserves error and metadata returned from `server.py`"""

    # WARN: extra ignore to ignore extra fields,
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="ignore")

    body: PydanticModel
    error: str | None = None
    metadata: dict[str, Any] = {}
    client_response: aiohttp.ClientResponse | None
    exception: BaseException | None = None

    @field_serializer("client_response")
    def serialize_client_response(
        self, value: aiohttp.ClientResponse | None
    ) -> dict[str, Any] | None:
        if value is None:
            return None
        return {
            "status": value.status,
            "headers": dict(value.headers),
            "url": str(value.url),
            "host": value.host,
        }

    @field_serializer("exception")
    def serialize_exception(self, value: BaseException | None) -> dict[str, Any] | None:
        if value is None:
            return None
        return {
            "type": type(value).__name__,
            "message": str(value),
            "traceback": traceback.format_exception(
                type(value), value, value.__traceback__
            ),
        }


class InterceptHandler(logging.Handler):
    def emit(self, record):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Use the logging record's information instead of trying to calculate depth
        try:
            logger.patch(
                lambda r: r.update(
                    name=record.name,
                    function=record.funcName,
                    file=record.pathname,
                    line=record.lineno,
                    module=record.module,
                )
            ).log(level, record.getMessage())
        except:  # noqa: E722
            pass



# from .server import _register_route_handler as _register_route_handler
from .client import Client, get_client
from .exceptions import InvalidSignatureException
from .middleware import SignatureMiddleware, ZstdMiddleware
from .server import Request, Server
from .types import HOTKEY_HEADER, PydanticModel, ServerHandlerFunc, StdResponse
from .utils import extract_headers

__all__ = [
    "Server",
    "Client",
    "get_client",
    "extract_headers",
    "StdResponse",
    "Request",
    "PydanticModel",
    "HOTKEY_HEADER",
    "ServerHandlerFunc",
    "ZstdMiddleware",
    "SignatureMiddleware",
    "InvalidSignatureException",
]



import asyncio
import http
from typing import Any, Sequence

import aiohttp
import orjson
import zstandard as zstd
from aiohttp.client import ClientSession
from kami import KamiClient
from loguru import logger
from orjson import JSONDecodeError
from pydantic import BaseModel
from tenacity import (
    AsyncRetrying,
    RetryError,
    stop_after_attempt,
    wait_exponential,
)

from .utils import retry_log

from .types import (
    HOTKEY_HEADER,
    MESSAGE_HEADER,
    SIGNATURE_HEADER,
    PydanticModel,
    StdResponse,
)
from .utils import encode_body


def get_client(conn_limit: int = None, limit_per_host: int = None) -> ClientSession:  # type: ignore[assignment]
    if not conn_limit:
        conn_limit = 256
    if not limit_per_host:
        limit_per_host = 10

    return aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(
            ssl=False,
            limit=conn_limit,
            limit_per_host=limit_per_host,
            enable_cleanup_closed=True,
        )
    )


def _build_url(url: str, model: BaseModel, protocol: str = "http") -> str:
    if not url.startswith("http") and not url.startswith("https"):
        url = f"{protocol}://{url}"

    return f"{url}/{model.__class__.__name__}"


async def _log_context(response: StdResponse[PydanticModel]) -> None:
    if response.exception:
        logger.error(f"Error due to exception: {response.exception}")
    elif response.error:
        logger.error(f"Error due to error: {response.error}")
    else:
        client_response = response.client_response
        if client_response and client_response.status != http.HTTPStatus.OK:
            logger.error(
                f"NOT OK, HTTP status: {client_response.status}, text: {await client_response.text()} error: {response.error if response else ''}, metadata: {response.metadata if response else ''}"
            )


class Client:
    def __init__(
        self,
        hotkey: str,
        session: ClientSession | None = None,
    ) -> None:
        self._kami = KamiClient()
        self._hotkey = hotkey
        self._session: ClientSession = session or get_client()
        self._compression_headers = {
            "content-encoding": "zstd",
            "accept-encoding": "zstd",
        }

    async def _build_headers(
        self,
        include_compression: bool = True,
        message: str = None,  # type: ignore[assignment]
    ) -> dict[str, str]:
        if not message:
            message: str = (
                f"I solemnly swear that I am up to some good. Hotkey: {self._hotkey}"
            )

        signature: str = await self._kami.sign_message(message)
        headers: dict[str, str] = {
            "content-type": "application/json",
            SIGNATURE_HEADER: signature,
            HOTKEY_HEADER: self._hotkey,
            MESSAGE_HEADER: message,
        }
        if include_compression:
            headers.update(self._compression_headers)
        logger.trace(f"Sending request with headers: {headers}")
        return headers

    async def _ensure_session(self):
        """Recreate session if it's closed"""
        if not self._session or self._session.closed:
            self._session = get_client()

    async def batch_send(
        self,
        urls: list[str],
        models: list[PydanticModel],
        semaphore: asyncio.BoundedSemaphore | None = None,
        **kwargs: Any,
    ) -> Sequence[StdResponse[PydanticModel]]:
        """Sends the following payloads to the given URLs concurrently.
        Expects that the endpoint is hosted at:
            http://<url>/<model_name> where model_name is the name of the Pydantic model

        Args:
            urls (list[str]): urls
            models (list[PydanticModel]): models
            keypair (substrateinterface.Keypair): keypair

        Returns:
            list[Response]: Returns both the aiohttp Response, and the model that
                was returned from the server, or the exception if the request failed
        """
        if semaphore is None:
            logger.info("Attempting to batch sending requests without semaphore")
            responses = await asyncio.gather(
                *[self.send(url, model, **kwargs) for url, model in zip(urls, models)],
            )
            for r in responses:
                await _log_context(r)

            return responses

        async def _send_with_semaphore(
            url: str, model: PydanticModel
        ) -> StdResponse[PydanticModel]:
            async with semaphore:
                return await self.send(url, model, **kwargs)

        responses = await asyncio.gather(
            *[_send_with_semaphore(url, model) for url, model in zip(urls, models)],
        )
        for r in responses:
            await _log_context(r)

        return responses

    async def send(
        self,
        url: str,
        model: PydanticModel,
        timeout_sec: int = 10,
        max_retries: int = 2,
        max_wait_sec: int = 4,
        wait_exponential_factor: int = 2,
        enable_preflight: bool = True,
        **kwargs: Any,
    ) -> StdResponse[PydanticModel]:
        """Sends the following payload to the given URL.
        Expects that the endpoint is hosted at:
            http://<url>/<model_name> where model_name is the name of the Pydantic model

        Args:
            url (str): url
            model (PydanticModel): model
            keypair (substrateinterface.Keypair): keypair
            max_retries (int): max number of retries
            max_wait_sec (int): max wait in unit of seconds

        Returns:
            Response: Returns both the aiohttp Response, and the model that
                was returned from the server
        """
        # NOTE: here we set some defaults to AT LEAST retry some
        model_name = model.__class__.__name__
        client_resp: aiohttp.ClientResponse | None = None
        context_msg = f"{url=}, {model_name=}, {max_retries=}, {max_wait_sec=}"
        try:
            await self._ensure_session()
            async for attempt in AsyncRetrying(
                stop=stop_after_attempt(max_retries),
                wait=wait_exponential(
                    multiplier=wait_exponential_factor, max=max_wait_sec
                ),
                before_sleep=retry_log,
            ):
                with attempt:
                    target_url = _build_url(url, model)

                    if enable_preflight:
                        _head_headers = await self._build_headers(
                            include_compression=False
                        )
                        async with self._session.head(
                            target_url,
                            headers=_head_headers,
                            timeout=aiohttp.ClientTimeout(total=timeout_sec),
                        ) as head_resp:
                            head_resp.raise_for_status()
                            logger.debug(f"HEAD preflight successful for {target_url}")

                    _headers = await self._build_headers()
                    payload = encode_body(model, _headers)
                    async with self._session.post(
                        target_url,
                        data=payload,
                        headers=_headers,
                        timeout=aiohttp.ClientTimeout(total=timeout_sec),
                    ) as client_resp:
                        # raise exception so we can retry
                        client_resp.raise_for_status()

                        logger.info(
                            f"Received response with status: {client_resp.status}, {context_msg}"
                        )
                        response_json = {}
                        try:
                            if (
                                client_resp.headers.get("content-encoding", "").lower()
                                == "zstd"
                            ):
                                logger.info(
                                    f"Attempting zstd decoding for {model_name}"
                                )
                                response_bytes = await client_resp.read()
                                dctx = zstd.ZstdDecompressor()
                                decompressed_bytes = dctx.decompress(response_bytes)
                                response_text = decompressed_bytes.decode()
                                response_json = orjson.loads(response_text)
                                logger.debug(
                                    f"Successfully parsed JSON: {response_json}, type: {type(response_json)}, {context_msg}"
                                )
                            else:
                                response_json = await client_resp.json()
                        except JSONDecodeError as e:
                            logger.error(
                                f"Failed to decode response: {await client_resp.text()}, {context_msg}, exception: {e}"
                            )
                            raise

                        if not response_json:
                            logger.warning("Empty response JSON received")
                            return StdResponse(
                                # NOTE: here we're creating an empty instance
                                body=model.model_construct(),
                                exception=ValueError(
                                    f"Empty response JSON received for {model_name}"
                                ),
                                client_response=client_resp,
                            )

                        try:
                            error: str | None = response_json.get("error", None)
                            metadata: dict[str, Any] = response_json.get("metadata", {})
                            body: dict[str, Any] = response_json.get("body", {})

                            if body:
                                try:
                                    # parse object to the specific model
                                    pydantic_model = model.model_validate(body)
                                    logger.success(
                                        f"Successfully received response, {context_msg}"
                                    )
                                    return StdResponse(
                                        body=pydantic_model,
                                        error=error,
                                        metadata=metadata,
                                        client_response=client_resp,
                                    )
                                except Exception as e:
                                    logger.error(
                                        f"Failed to validate model with body: {e}, returning the raw body"
                                    )
                                    # Return the raw body if validation fails
                                    return StdResponse(
                                        body=model.model_construct(**body),
                                        error=error,
                                        metadata=metadata,
                                        client_response=client_resp,
                                    )
                            else:
                                logger.warning(
                                    "Response body is empty, returning empty model"
                                )
                                return StdResponse(
                                    body=model.model_construct(),
                                    error=error,
                                    metadata=metadata,
                                    client_response=client_resp,
                                )

                        except JSONDecodeError as e:
                            logger.error(f"Failed to decode response: {e}")
                            raise

            return StdResponse(
                body=model.model_construct(),
                exception=None,
                client_response=client_resp,
            )
        except asyncio.CancelledError:
            logger.warning(f"Request to {url} was cancelled but may still be running")
            # You can choose to return a response or re-raise
            return StdResponse(
                body=model.model_construct(),
                exception=asyncio.CancelledError("Request was cancelled"),
                client_response=client_resp,
            )
        except KeyboardInterrupt:
            logger.warning(f"KeyboardInterrupt during request to {url}")
            return StdResponse(
                body=model.model_construct(),
                exception=KeyboardInterrupt("Request interrupted"),
                client_response=client_resp,
            )

        except RetryError as e:
            logger.error(
                f"All retries to {url} for {model_name} with {max_retries=}, {max_wait_sec=} were exhausted"
            )
            logger.error(f"Final exception: {e.last_attempt.exception()}")
            return StdResponse(
                body=model.model_construct(), exception=e, client_response=client_resp
            )
        except (Exception, BaseException) as e:
            return StdResponse(
                body=model.model_construct(), exception=e, client_response=client_resp
            )

    async def close(self):
        try:
            await self._session.close()
        except Exception:
            pass



from typing import Any


from loguru import logger
from tenacity import (
    RetryCallState,
)

import zstandard as zstd
from fastapi import HTTPException, Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import ORJSONResponse
from pydantic import BaseModel

from .types import HOTKEY_HEADER, MESSAGE_HEADER, SIGNATURE_HEADER


def create_response(
    body: dict[str, Any],
    status_code: int = 200,
    error: str | None = None,
    metadata: dict[str, Any] = {},
):
    """
    Helper function to create standardized RESTful API responses

    Args:
        body: The response data
        status_code: HTTP status code (default: 200)
        error: Optional error message for error responses
        metadata: Optional metadata like pagination info, request ID, etc.
    """
    content = {"body": jsonable_encoder(body), "error": error, "metadata": {}}  # pyright: ignore

    if metadata:
        content["metadata"] = jsonable_encoder(metadata)

    return ORJSONResponse(content=content, status_code=status_code)


def encode_body(model: BaseModel, headers: dict[str, Any]) -> bytes:
    content_encoding = headers.get("content-encoding")
    if content_encoding:
        if content_encoding.lower() == "zstd":
            json_data = model.model_dump_json().encode()
            compressor = zstd.ZstdCompressor(level=3)
            compressed = compressor.compress(json_data)
            return compressed
        else:
            raise NotImplementedError(
                f"Content encoding of type {content_encoding} is not supported at the moment"
            )

    return model.model_dump_json().encode()


async def decode_body(request: Request) -> bytes:
    """Handle zstd decoding to make transmission over network smaller"""
    body = await request.body()
    if (
        "content-encoding" in request.headers
        and "zstd" in request.headers["content-encoding"]
    ):
        try:
            decompressor = zstd.ZstdDecompressor()
            body = decompressor.decompress(body)
        except Exception as e:
            raise HTTPException(
                status_code=400, detail=f"Failed to decompress zstd data: {str(e)}"
            )

    return body


def extract_headers(request: Request) -> tuple[str, str, str]:
    """Based on the headers, extract the hotkey, message and signature"""
    try:
        headers: dict[str, Any] = {}
        for header, value in request.headers.items():
            if header.startswith("X-"):
                headers[header] = value
        signature = headers.get(SIGNATURE_HEADER, "")
        hotkey = headers.get(HOTKEY_HEADER, "")
        message = headers.get(MESSAGE_HEADER, "")
        return hotkey, message, signature

    except Exception as e:
        logger.warning(f"Failed to extract_headers: {e}")
        return "", "", ""


def retry_log(retry_state: RetryCallState):
    """Custom retry logger that works well with loguru"""
    func_name = getattr(retry_state.fn, "__name__", "<unknown_function>")
    logger.debug(
        f"Retrying {func_name} attempt {retry_state.attempt_number} "
        f"after {retry_state.seconds_since_start:.1f}s due to: {retry_state.outcome.exception() if retry_state.outcome else ''}"
    )



from http import HTTPStatus

from aiohttp import ClientSession, TCPConnector
from pydantic import BaseModel, Field
from typing import Any

from .client import Client


# retry logic handled under the hood
class ExampleModel(BaseModel):
    field: bool = Field(default=False, description="Example field")


async def main():
    # provide a session is optional, or use the convenience function
    client = Client(
        hotkey="your hotkey",  # used for signing and verification,
        session=ClientSession(
            connector=TCPConnector(
                limit=100
            )  # specify your own session, connector, etc. as needed
        ),
    )

    example_model = ExampleModel()
    # any other kwargs may be propagated to the underlying POST request using aiohttp
    kwargs: dict[str, Any] = {}
    response = await client.send(
        "http://ip:port",
        model=example_model,
        timeout_sec=12,
        max_retries=10,
        max_wait_sec=100,
        **kwargs,
    )

    # check for errors
    if response.error:
        # check for response.error that comes from server side
        print(f"Error received from server: {response.error}")
    if response.exception:
        # check for response.exception that comes from client side
        print(f"Exception occurred {response.exception}")

    if response.client_response and response.client_response.status != HTTPStatus.OK:
        # access fields from ExampleModel directly
        print(f"Received {response.body.field=}")

    # dedicated batch_send method
    # use an asyncio semaphore for concurrency & batch control
    # if semaphore isn't provided, asyncio.gather is called
    responses = await client.batch_send(
        urls=["http://ip:port"] * 5, models=[ExampleModel()] * 5
    )
    for response in responses:
        # access fields as needed...

        # check for errors
        if response.error:
            # check for response.error that comes from server side
            print(f"Error received from server: {response.error}")
        if response.exception:
            # check for response.exception that comes from client side
            print(f"Exception occurred {response.exception}")

        if (
            response.client_response
            and response.client_response.status != HTTPStatus.OK
        ):
            # access fields from ExampleModel directly
            print(f"Received {response.body.field=}")



import http
from typing import Awaitable, Callable

from starlette.types import ASGIApp
import zstandard as zstd
from fastapi import Request, Response
from fastapi.responses import ORJSONResponse
from kami import KamiClient
from loguru import logger
from starlette.concurrency import iterate_in_threadpool
from starlette.middleware.base import BaseHTTPMiddleware

from .types import HOTKEY_HEADER, MESSAGE_HEADER, SIGNATURE_HEADER
from .utils import (
    create_response,
    decode_body,
)

compressor = zstd.ZstdCompressor(level=3)


class SignatureMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        kami: KamiClient,
        whitelisted_routes: list[str] | None = None,
    ):
        super().__init__(app)
        self.kami = kami
        self.whitelisted_routes = whitelisted_routes or []
        # always whitelisted
        if "/docs" not in self.whitelisted_routes:
            self.whitelisted_routes.append("/docs")

    async def dispatch(
        self, request: Request, call_next: Callable[[Request], Awaitable[Response]]
    ) -> ORJSONResponse | Response:
        if request.url.path in self.whitelisted_routes:
            return await call_next(request)

        signature = request.headers.get(SIGNATURE_HEADER, "")
        hotkey = request.headers.get(HOTKEY_HEADER, "")
        message = request.headers.get(MESSAGE_HEADER, "")
        if not hotkey or not signature or not message:
            message = f"{http.HTTPStatus(400).phrase}, missing \
                    headers, expected: {SIGNATURE_HEADER}, {HOTKEY_HEADER}, {MESSAGE_HEADER}, \
                    got: {hotkey=}, {signature=}, {message=}"
            return create_response(body={}, status_code=400, error=message)

        if not await self.kami.verify(
            hotkey=hotkey, message=message, signature=signature
        ):
            return create_response(
                body={},
                status_code=403,
                error=f"{http.HTTPStatus(403).phrase} due to invalid signature",
            )

        # otherwise, proceed with the normal request
        response = await call_next(request)
        return response


class ZstdMiddleware(BaseHTTPMiddleware):
    """Middleware that handles zstd compression/decompression for request and response bodies.

    This middleware:
    1. Decompresses incoming request bodies with content-encoding: zstd
    2. Compresses outgoing response bodies when Accept-Encoding includes zstd

    NOTE: The /docs endpoint is excluded from compression/decompression.
    """

    def __init__(self, app: ASGIApp, whitelisted_routes: list[str] | None = None):
        super().__init__(app)
        self.whitelisted_routes = whitelisted_routes or []
        # always whitelisted
        if "/docs" not in self.whitelisted_routes:
            self.whitelisted_routes.append("/docs")

    async def dispatch(
        self, request: Request, call_next: Callable[[Request], Awaitable[Response]]
    ) -> ORJSONResponse | Response:
        logger.debug(f"Request headers: {request.headers}")
        if request.url.path in self.whitelisted_routes:
            return await call_next(request)

        # Skip compression/decompression for HEAD requests (no body to process)
        if request.method == "HEAD":
            return await call_next(request)

        encoding = request.headers.get("content-encoding", "").lower()
        if encoding == "zstd":
            decompressed_body = await decode_body(request)
            # NOTE: this probably won't work for streaming
            request._body = decompressed_body  # pyright: ignore[reportPrivateUsage]
            logger.debug("Server decompressed request body")

        # Process the request
        response = await call_next(request)
        logger.debug(f"raw response: {response=}")

        response_body = [section async for section in response.body_iterator]  # type: ignore
        response.body_iterator = iterate_in_threadpool(iter(response_body))  # type: ignore
        bytes_response = response_body[0]  # type: ignore
        logger.debug(f"response_body={bytes_response.decode()}")

        accept_encoding = request.headers.get("accept-encoding", "").lower()
        if response_body and "zstd" in accept_encoding:
            compressed_body = compressor.compress(bytes_response)

            new_response = Response(
                content=compressed_body,
                status_code=response.status_code,
                headers=dict(response.headers),
            )

            new_response.headers["content-encoding"] = "zstd"
            new_response.headers["content-length"] = str(len(compressed_body))

            logger.debug(
                f"Compressed response: original_size={len(bytes_response)}, compressed_size={len(compressed_body)}"
            )
            return new_response
        else:
            if response_body:
                response_content = response_body[0]
                response.headers["content-length"] = str(len(response_content))

            logger.debug(
                f"Not compressing response: accept_encoding={accept_encoding}, has_body={bool(response_body)}, body_size={len(response_body[0]) if response_body else 0}"
            )

        return response
